@@ -25,16 +25,18 @@
 
 #include <string.h>
 
-#include "e-util/e-path.h"
 #include <camel/camel-i18n.h>
 
 #include "camel-exchange-store.h"
 #include "camel-exchange-folder.h"
 #include <camel/camel-session.h>
 
-#include <gal/util/e-util.h>
+#define SUBFOLDER_DIR_NAME     "subfolders"
+#define SUBFOLDER_DIR_NAME_LEN 10
 
-static CamelStoreClass *parent_class = NULL;
+
+//static CamelStoreClass *parent_class = NULL;
+static CamelOfflineStoreClass *parent_class = NULL;
 
 #define CS_CLASS(so) ((CamelStoreClass *)((CamelObject *)(so))->klass)
 
@@ -47,6 +49,7 @@ static char  *get_name         (CamelService *service, gboolean brief);
 static CamelFolder     *get_trash       (CamelStore *store,
 					 CamelException *ex);
 
+char * exchange_path_to_physical (const char *prefix, const char *vpath);
 static gboolean exchange_connect (CamelService *service, CamelException *ex);
 static gboolean exchange_disconnect (CamelService *service, gboolean clean, CamelException *ex);
 
@@ -78,7 +81,7 @@ class_init (CamelExchangeStoreClass *camel_exchange_store_class)
 	CamelStoreClass *camel_store_class =
 		CAMEL_STORE_CLASS (camel_exchange_store_class);
 	
-	parent_class = CAMEL_STORE_CLASS (camel_type_get_global_classfuncs (camel_store_get_type ()));
+	parent_class = CAMEL_OFFLINE_STORE_CLASS (camel_type_get_global_classfuncs (camel_offline_store_get_type ()));
 	
 	/* virtual method overload */
 	camel_service_class->construct = construct;
@@ -125,7 +128,7 @@ camel_exchange_store_get_type (void)
 
 	if (!camel_exchange_store_type) {
 		camel_exchange_store_type = camel_type_register (
-			CAMEL_STORE_TYPE,
+			camel_offline_store_get_type (),
 			"CamelExchangeStore",
 			sizeof (CamelExchangeStore),
 			sizeof (CamelExchangeStoreClass),
@@ -138,6 +141,84 @@ camel_exchange_store_get_type (void)
 	return camel_exchange_store_type;
 }
 
+/* This has been now removed from evolution/e-util. So implemented this here.
+ * Also note that this is similar to the call in e2k-path.c. The name of the 
+ * function has been changed to avoid any conflicts.
+ */
+char *
+exchange_path_to_physical (const char *prefix, const char *vpath)
+{
+	const char *p, *newp;
+	char *dp;
+	char *ppath;
+	int ppath_len;
+	int prefix_len;
+
+	while (*vpath == '/')
+		vpath++;
+	if (!prefix)
+		prefix = "";
+
+	/* Calculate the length of the real path. */
+	ppath_len = strlen (vpath);
+	ppath_len++;	/* For the ending zero.  */
+
+	prefix_len = strlen (prefix);
+	ppath_len += prefix_len;
+	ppath_len++;	/* For the separating slash.  */
+
+	/* Take account of the fact that we need to translate every
+	 * separator into `subfolders/'.
+	 */
+	p = vpath;
+	while (1) {
+		newp = strchr (p, '/');
+		if (newp == NULL)
+			break;
+
+		ppath_len += SUBFOLDER_DIR_NAME_LEN;
+		ppath_len++; /* For the separating slash.  */
+
+		/* Skip consecutive slashes.  */
+		while (*newp == '/')
+			newp++;
+
+		p = newp;
+	};
+
+	ppath = g_malloc (ppath_len);
+	dp = ppath;
+
+	memcpy (dp, prefix, prefix_len);
+	dp += prefix_len;
+	*(dp++) = '/';
+
+	/* Copy the mangled path.  */
+	p = vpath;
+ 	while (1) {
+		newp = strchr (p, '/');
+		if (newp == NULL) {
+			strcpy (dp, p);
+			break;
+		}
+
+		memcpy (dp, p, newp - p + 1); /* `+ 1' to copy the slash too.  */
+		dp += newp - p + 1;
+
+		memcpy (dp, SUBFOLDER_DIR_NAME, SUBFOLDER_DIR_NAME_LEN);
+		dp += SUBFOLDER_DIR_NAME_LEN;
+
+		*(dp++) = '/';
+
+		/* Skip consecutive slashes.  */
+		while (*newp == '/')
+			newp++;
+
+		p = newp;
+	}
+
+	return ppath;
+}
 static void
 construct (CamelService *service, CamelSession *session,
 	   CamelProvider *provider, CamelURL *url, CamelException *ex)
@@ -208,7 +289,7 @@ exchange_connect (CamelService *service, CamelException *ex)
 	
 	store->stub = camel_stub_new (socket_path, _("Evolution Exchange backend process"), ex);
 	g_free (socket_path);
-	if (!store->stub) 
+	if (!store->stub)
 		return FALSE;
 
 	/* Initialize the stub connection */
@@ -225,7 +306,7 @@ exchange_connect (CamelService *service, CamelException *ex)
 
 	camel_object_hook_event (CAMEL_OBJECT (store->stub), "notification",
 				 stub_notification, store);
-	
+
 	return TRUE;
 }
 
@@ -251,14 +332,25 @@ exchange_disconnect (CamelService *service, gboolean clean, CamelException *ex)
 
 static CamelFolder *
 exchange_get_folder (CamelStore *store, const char *folder_name,
-		     guint32 camel_flags, CamelException *ex)
+		     guint32 flags, CamelException *ex)
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 	CamelFolder *folder;
 	char *folder_dir;
-
+	
 	RETURN_VAL_IF_NOT_CONNECTED (store, ex, NULL);
-
+	
+	folder_dir = exchange_path_to_physical (exch->storage_path, folder_name);
+	
+	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		if (!folder_dir || access (folder_dir, F_OK) != 0) {
+			g_free (folder_dir);
+			camel_exception_setv (ex, CAMEL_EXCEPTION_STORE_NO_FOLDER,
+					      _("No such folder %s"), folder_name);
+			return NULL;
+		}
+	}
+	
 	g_mutex_lock (exch->folders_lock);
 	folder = g_hash_table_lookup (exch->folders, folder_name);
 	if (folder) {
@@ -267,16 +359,16 @@ exchange_get_folder (CamelStore *store, const char *folder_name,
 		 */
 		g_mutex_unlock (exch->folders_lock);
 		camel_object_ref (CAMEL_OBJECT (folder));
+		g_free (folder_dir);
 		return folder;
 	}
 
 	folder = (CamelFolder *)camel_object_new (CAMEL_EXCHANGE_FOLDER_TYPE);
 	g_hash_table_insert (exch->folders, g_strdup (folder_name), folder);
 	g_mutex_unlock (exch->folders_lock);
-
-	folder_dir = e_path_to_physical (exch->storage_path, folder_name);
+	
 	if (!camel_exchange_folder_construct (folder, store, folder_name,
-					      camel_flags, folder_dir,
+					      flags, folder_dir, ((CamelOfflineStore *) store)->state,
 					      exch->stub, ex)) {
 		g_free (folder_dir);
 		camel_object_unref (CAMEL_OBJECT (folder));
@@ -335,6 +427,12 @@ make_folder_info (CamelExchangeStore *exch, char *name, char *uri,
 	if (flags & CAMEL_STUB_FOLDER_NOSELECT)
 		info->flags = CAMEL_FOLDER_NOSELECT;
 
+	if (flags & CAMEL_STUB_FOLDER_SYSTEM)
+		info->flags |= CAMEL_FOLDER_SYSTEM;
+		
+	if (flags & CAMEL_STUB_FOLDER_TYPE_INBOX)
+		info->flags |= CAMEL_FOLDER_TYPE_INBOX;
+
 	return info;
 }
 
@@ -366,8 +464,7 @@ postprocess_tree (CamelFolderInfo *info)
 			
 
 static CamelFolderInfo *
-exchange_get_folder_info (CamelStore *store, const char *top,
-			  guint32 flags, CamelException *ex)
+exchange_get_folder_info (CamelStore *store, const char *top, guint32 flags, CamelException *ex)
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 	GPtrArray *folders, *folder_names, *folder_uris;
@@ -375,7 +472,12 @@ exchange_get_folder_info (CamelStore *store, const char *top,
 	GByteArray *folder_flags;
 	CamelFolderInfo *info;
 	int i;
-
+#if 0	
+	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot get folder info in offline mode."));
+		return NULL;
+	}
+#endif	
 	RETURN_VAL_IF_NOT_CONNECTED (store, ex, NULL);
 
 	/* If the backend crashed, don't keep returning an error
@@ -432,6 +534,11 @@ exchange_create_folder (CamelStore *store, const char *parent_name,
 	char *folder_uri;
 	guint32 unread_count, flags;
 	CamelFolderInfo *info;
+	
+	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot create folder in offline mode."));
+		return NULL;
+	}
 
 	if (!camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_CREATE_FOLDER,
 			      CAMEL_STUB_ARG_FOLDER, parent_name,
@@ -455,6 +562,11 @@ exchange_delete_folder (CamelStore *store, const char *folder_name,
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 
+	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot delete folder in offline mode."));
+		return;
+	}
+
 	camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_DELETE_FOLDER,
 			 CAMEL_STUB_ARG_FOLDER, folder_name,
 			 CAMEL_STUB_ARG_END);
@@ -466,6 +578,11 @@ exchange_rename_folder (CamelStore *store, const char *old_name,
 {
 	CamelExchangeStore *exch = CAMEL_EXCHANGE_STORE (store);
 
+	if (((CamelOfflineStore *) store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		camel_exception_set (ex, CAMEL_EXCEPTION_SYSTEM, _("Cannot rename folder in offline mode."));
+		return;
+	}
+
 	camel_stub_send (exch->stub, ex, CAMEL_STUB_CMD_RENAME_FOLDER,
 			 CAMEL_STUB_ARG_FOLDER, old_name,
 			 CAMEL_STUB_ARG_FOLDER, new_name,
